import { __decorate } from "tslib";
import { createErrorMap } from '@genesislcap/foundation-utils';
import { defaultExecutionContext, observable, Observable, } from '@microsoft/fast-element';
import { Container } from '@microsoft/fast-foundation';
import { interpret, InterpreterStatus } from 'xstate';
import { logger } from '../utils';
import { getBindArgs } from './binding';
/**
 * AbstractMachine base class.
 * @public
 */
export class AbstractMachine {
    constructor() {
        /** {@inheritDoc Machine.onSendEvent} */
        this.onSendEvent = (event) => {
            this.send(event.detail);
        };
    }
    /** {@inheritDoc Machine.start} */
    start(options) {
        this.interpret(options);
        this.subscribe();
        this.startActor();
    }
    /** {@inheritDoc Machine.stop} */
    stop() {
        this.stopActor();
        this.unsubscribe();
    }
    /**
     * @internal
     */
    interpret(options = {}) {
        if (!this.machine) {
            throw new Error('Create this.machine before calling interpret()');
        }
        const contextWithErrors = {
            errors: createErrorMap(logger.error),
            error: undefined,
        };
        this.startingState = options === null || options === void 0 ? void 0 : options.state;
        this.startingContext = (options.input
            ? Object.assign(Object.assign({}, contextWithErrors), options.input) : contextWithErrors);
        this.actor = interpret(this.machine, Object.assign(Object.assign({}, options), { input: this.startingContext, state: this.startingState }));
        const { value: currentValue, context: currentContext } = this.actor.getSnapshot();
        this.state = currentValue;
        this.context = currentContext;
        this.complete = false;
        this.output = undefined;
        this.errors = currentContext.errors;
        this.error = undefined;
    }
    /**
     * @internal
     */
    subscribe() {
        if (!this.actor) {
            throw new Error('Call interpret() before calling subscribe()');
        }
        this.subscription = this.actor.subscribe({
            next: (snapshot) => {
                var _a;
                this.state = snapshot.value;
                this.context = snapshot.context;
                this.errors = this.context.errors;
                this.error = (_a = this.context.error) !== null && _a !== void 0 ? _a : this.errors.lastError;
                logger.debug(`${this.machine.id} changed, state:`, this.state, `context:`, this.context);
            },
            error: (err) => {
                this.error = err;
                logger.error(`${this.machine.id} error:`, err);
            },
            complete: () => {
                this.complete = true;
                this.output = this.actor.getSnapshot().output;
                logger.debug(`${this.machine.id} complete:`, this.output);
            },
        });
    }
    /**
     * @internal
     */
    unsubscribe() {
        this.subscription && this.subscription.unsubscribe();
    }
    /**
     * @internal
     */
    startActor() {
        var _a;
        if (((_a = this.actor) === null || _a === void 0 ? void 0 : _a.status) !== InterpreterStatus.Running) {
            this.actor.start();
        }
    }
    /**
     * @internal
     */
    stopActor() {
        var _a;
        if (((_a = this.actor) === null || _a === void 0 ? void 0 : _a.status) === InterpreterStatus.Running) {
            this.actor.stop();
        }
    }
    /** {@inheritDoc Machine.provide} */
    provide(implementations) {
        if (!this.machine) {
            throw new Error('Create this.machine before calling provide()');
        }
        this.machine = this.machine.provide(implementations);
        if (this.actor) {
            this.stop();
            logger.debug('You will need to recall start() after calling provide() on a running machine.');
        }
    }
    /** {@inheritDoc Machine.matches} */
    matches(value) {
        if (!this.actor) {
            throw new Error('Call interpret() before calling matches()');
        }
        return this.actor.getSnapshot().matches(value);
    }
    /** {@inheritDoc Machine.hasTag} */
    hasTag(tag) {
        if (!this.actor) {
            throw new Error('Call interpret() before calling hasTag()');
        }
        return this.actor.getSnapshot().hasTag(tag);
    }
    /** {@inheritDoc Machine.send} */
    send(event) {
        this.actor.send(event);
    }
    /** {@inheritDoc Machine.binding} */
    binding(token, subscriberChangeCallback, isVolatileBinding = false, context = defaultExecutionContext) {
        const valueGetter = getBindArgs(token);
        const initialSubscriber = {
            handleChange: (binding, args) => {
                if (subscriberChangeCallback) {
                    subscriberChangeCallback(binding(this, context), args);
                }
            },
        };
        const bindingObserver = Observable.binding(valueGetter, initialSubscriber, isVolatileBinding);
        bindingObserver.observe(this, context);
        return bindingObserver;
    }
}
__decorate([
    Container
], AbstractMachine.prototype, "container", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "machine", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "actor", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "state", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "startingState", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "context", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "startingContext", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "complete", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "output", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "errors", void 0);
__decorate([
    observable
], AbstractMachine.prototype, "error", void 0);
